/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.vew.daten;

import de.bsvrz.dav.daf.main.Data;

/**
 * Interface zur Konvertierung DAV-Daten in externe Daten und umgekehrt.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: ExterneVerkehrsmeldungenAdapterIf.java 63753 2016-10-06 09:44:08Z gieseler $
 */

public interface ExterneVerkehrsmeldungenAdapterIf {

	/**
	 * Konvertiert DAV-Daten in externe Daten.
	 * 
	 * @param davdata
	 *            {@link de.bsvrz.dav.daf.main.Data}
	 * @param id
	 *            zu benutzende Meldungs-ID
	 * @return {@link ExterneMeldungsDaten}
	 * @throws TMCKonvertierungException
	 *             bei Fehlern
	 */
	ExterneMeldungsDaten erzeugeExterneMeldungsdaten(Data davdata, final String id) throws TMCKonvertierungException;

	/**
	 * Konvertiert die Daten in DAV-Daten.
	 * 
	 * @param davdata
	 *            {@link de.bsvrz.dav.daf.main.Data}
	 * @throws TMCKonvertierungException
	 *             bei Fehlern
	 */
	void erzeugeDAVDaten(Data davdata) throws TMCKonvertierungException;

	/**
	 * Pr&uuml;ft, ob die Meldung TMC-Daten besitzt.
	 * <p>
	 * Das Kriterium ist, dass die Meldung eine Verortung über TMC-Lokationen
	 * oder Koordinaten besitzt, damit sie dargestellt werden kann.
	 * </p>
	 * <p>
	 * Fehlende TMC-Ereigniscodes werden hier nicht herangezogen, da das
	 * Ereignis auch im Meldungstext enthalten ist.
	 * </p>
	 * 
	 * @return <code>true</code> wenn TMC_daten vorhanden, sonst
	 *         <code>false</code>
	 */
	boolean hasTmcDaten();
}
