/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.vew.daten.tic;

import java.util.ArrayList;
import java.util.List;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.kex.lmstic3.vew.LMSTIC3LogTools;
import de.bsvrz.kex.lmstic3.vew.LmsMsg;
import de.bsvrz.kex.lmstic3.vew.daten.ExterneMeldungsDaten;
import de.bsvrz.kex.lmstic3.vew.daten.ExterneVerkehrsmeldungenAdapterIf;
import de.bsvrz.kex.lmstic3.vew.daten.TMCKonvertierungException;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCEvent;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCEvent.TMCEventType;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCException;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCKoordinate;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCLocation;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCPrioritaet.TMCPrioritaetEnum;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCRichtung.TMCRichtungEnum;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCStatus.TMCStatusEnum;
import de.bsvrz.kex.lmstic3.vew.daten.tmc.TMCZustand;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * TIC3-Converter.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: TIC3Converter.java 64047 2016-10-27 15:42:01Z gieseler $
 */
public class TIC3Converter implements ExterneVerkehrsmeldungenAdapterIf {

	private ExterneMeldungsDaten daten;

	/**
	 * Konstruktor.
	 * 
	 * @param meldungsDaten
	 *            {@link ExterneMeldungsDaten}
	 */
	public TIC3Converter(final ExterneMeldungsDaten meldungsDaten) {
		super();
		this.daten = meldungsDaten;
	}

	@Override
	public ExterneMeldungsDaten erzeugeExterneMeldungsdaten(final Data davdata, final String id)
			throws TMCKonvertierungException {
		final ExterneMeldungsDaten tic3daten = new TIC3MeldungsDaten();

		tic3daten.setID(id);

		Data item = davdata.getItem("TMCID");
		tic3daten.setDatenKennung(item.getTextValue("Datenkennung").getText());
		tic3daten.setDatenProduzent(item.getTextValue("Datenproduzent").getText());

		tic3daten.setEreignisBeschreibung(davdata.getTextValue("Beschreibung").getText());

		item = davdata.getItem("TMCDaten");

		final Data vi = item.getItem("VerwaltungsInformationen");

		tic3daten.setDatenErzeugungsZeit(getZeitStempel(vi, "ErzeugungsZeit"));
		tic3daten.setDatenAktualisierungsZeit(getZeitStempel(vi, "AktualisierungsZeit"));
		tic3daten.setDatenStartZeit(getZeitStempel(vi, "StartZeit"));
		tic3daten.setDatenEndeZeit(getZeitStempel(vi, "EndeZeit"));

		final TMCStatusEnum stat;
		try {
			stat = TMCStatusEnum.getInstance(vi.getItem("TmcStatus").asUnscaledValue().intValue());

			tic3daten.setTMCMeldungsStatus(stat);

			final TMCPrioritaetEnum prio = TMCPrioritaetEnum
					.getInstance(vi.getItem("TmcPrioritaet").asUnscaledValue().intValue());
			tic3daten.setTMCMeldungsPrioritaet(prio);

			// Hier muss die OppositeDirectionLocation versorgt werden,
			// dann
			// gelten die Daten für beide Richtungen
			// final boolean istEreignisInBeidenRichtungen =
			// vi.getItem("EreignisInBeidenRichtungen").asUnscaledValue().intValue()
			// == 1;
			// tic3daten.setEreignisInBeidenRichtungen(istEreignisInBeidenRichtungen);
			// final boolean istEreignisInBeidenRichtungen =
			// tic3daten.istEreignisInBeidenRichtungen();

			convertEreignisse(davdata, tic3daten);
			convertVerortung(davdata, tic3daten);
		} catch (final TIC3Exception e) {
			throw new TMCKonvertierungException(e.getMessage());
		}

		return tic3daten;
	}

	private Long getZeitStempel(final Data data, final String item) {
		final Array array = data.getArray(item);
		if (array.getLength() > 0) {
			return array.getItem(0).asTimeValue().getMillis();
		}

		return null;
	}

	private void convertEreignisse(final Data davdata, final ExterneMeldungsDaten tic3daten) throws TIC3Exception {
		final Data item = davdata.getItem("TMCDaten");

		final Data ei = item.getItem("EreignisInformationen");
		final Array array = ei.getArray("EreignisDaten");

		if (array.getLength() == 0) {
			return;
		}

		final List<TMCEvent> tmcEreignisse = new ArrayList<>();

		for (int i = 0; i < array.getLength(); i++) {
			final Data ereignis = array.getItem(i);
			final int ecode = ereignis.getItem("EreignisCode").asUnscaledValue().intValue();
			final int empfcode = ereignis.getItem("EmpfehlungsCode").asUnscaledValue().intValue();
			final int vcode = ereignis.getItem("VorhersageCode").asUnscaledValue().intValue();

			// die internen Ereignis- und Vorhersagecodes werden beide auf
			// TIC-Eriegniscodes abgebildet
			if (ecode != 0) {
				tmcEreignisse.add(new TMCEvent(TMCEventType.Ereigniscode, ecode));
			}
			if (vcode != 0) {
				tmcEreignisse.add(new TMCEvent(TMCEventType.Ereigniscode, vcode));
			}
			if (empfcode != 0) {
				tmcEreignisse.add(new TMCEvent(TMCEventType.Empfehlungscode, empfcode));
			}
		}

		tic3daten.setTMCEreignisse(tmcEreignisse);
	}

	private void convertVerortung(final Data davdata, final ExterneMeldungsDaten tic3daten) throws TIC3Exception {
		final Data item = davdata.getItem("TMCDaten");

		final Data ei = item.getItem("VerortungsInformationen");
		final Array array = ei.getArray("LokationsDaten");

		if (array.getLength() == 0) {
			return;
		}

		final List<TMCLocation> lokationen = new ArrayList<>();

		for (int i = 0; i < array.getLength(); i++) {
			final Data loc = array.getItem(i);
			Double startOffset = null;
			if (!loc.getItem("StartOffset").asUnscaledValue().isState()) {
				// IntegerValueState state2 =
				// loc.getItem("StartOffset").asUnscaledValue().getState();
				// String name = state2.getName();
				startOffset = loc.getItem("StartOffset").asScaledValue().doubleValue();
			}

			Double endOffset = null;
			if (!loc.getItem("EndOffset").asUnscaledValue().isState()) {
				endOffset = loc.getItem("EndOffset").asScaledValue().doubleValue();
			}

			final int prim = loc.getItem("PrimaerLokation").asUnscaledValue().intValue();
			final int sek = loc.getItem("SekundaerLokation").asUnscaledValue().intValue();

			final TMCRichtungEnum r = TMCRichtungEnum.Positiv;

			final TMCLocation tl = new TMCLocation(prim, sek, startOffset, endOffset, r);
			lokationen.add(tl);
		}

		tic3daten.setTMCLokationen(lokationen);
	}

	@Override
	public void erzeugeDAVDaten(final Data davdata) throws TMCKonvertierungException {

		Data item = davdata.getItem("TMCID");
		item.getTextValue("Datenproduzent").setText(daten.getDatenProduzent());
		item.getTextValue("Datenkennung").setText(daten.getDatenKennung());

		davdata.getTextValue("Beschreibung").setText(daten.getEreignisBeschreibung());

		item = davdata.getItem("TMCDaten");

		final Data vi = item.getItem("VerwaltungsInformationen");

		try {
			daten.getDatenErzeugungsZeit().setDavDaten(vi, "ErzeugungsZeit");
			daten.getDatenAktualisierungsZeit().setDavDaten(vi, "AktualisierungsZeit");
			daten.getDatenStartZeit().setDavDaten(vi, "StartZeit");
			daten.getDatenEndeZeit().setDavDaten(vi, "EndeZeit");
		} catch (final TIC3Exception e) {
			throw new TMCKonvertierungException(e.getMessage());
		}

		final TMCStatusEnum stat = daten.getTMCMeldungsStatus();
		vi.getItem("TmcStatus").asUnscaledValue().set(stat.getId());

		final TMCPrioritaetEnum prio = daten.getTMCMeldungsPrioritaet();
		vi.getItem("TmcPrioritaet").asUnscaledValue().set(prio.getId());

		vi.getItem("Erzeugungsart").asTextValue().setText("LMS");

		final boolean istEreignisInBeidenRichtungen = daten.istEreignisInBeidenRichtungen();

		vi.getItem("EreignisInBeidenRichtungen").asUnscaledValue().set(istEreignisInBeidenRichtungen ? 1 : 0);

		convertEreignisseToDav(item, daten);
		convertVerortungToDav(item, daten);
	}

	private void convertEreignisseToDav(final Data item, final ExterneMeldungsDaten tic3daten) {
		final Data ei = item.getItem("EreignisInformationen");

		final List<TMCEvent> tmcEreignisse = tic3daten.getTMCEreignisse();

		final Array array = ei.getArray("EreignisDaten");

		final List<TMCZustand> okliste = new ArrayList<>();

		// Ereignisdaten testweise publizieren, Fehler werden geloggt, nur
		// fehlerfreie Daten werden weiter genutzt
		if (tmcEreignisse != null) {
			final int anzahl = tmcEreignisse.size();
			array.setLength(anzahl);
			for (int i = 0; i < anzahl; i++) {
				final TMCEvent te = tmcEreignisse.get(i);
				final TMCZustand ev = getDavEreignisKodierung(te, array.getItem(0));
				if (ev != null) {
					okliste.add(ev);
				}
			}
		}

		final int anzahl = okliste.size();
		array.setLength(anzahl);

		final TMCZustand ereignisCodeUndefiniert = new TMCZustand("EreignisCode", 0);
		final TMCZustand empfehlungsCodeCodeUndefiniert = new TMCZustand("EmpfehlungsCode", 0);
		final TMCZustand vorhersageCodeUndefiniert = new TMCZustand("VorhersageCode", 0);

		for (int i = 0; i < anzahl; i++) {
			try {
				final Data eitem = array.getItem(i);
				ereignisCodeUndefiniert.setDavDaten(eitem);
				empfehlungsCodeCodeUndefiniert.setDavDaten(eitem);
				vorhersageCodeUndefiniert.setDavDaten(eitem);

				final TMCZustand ev = okliste.get(i);
				ev.setDavDaten(eitem);
			} catch (final TMCException rdse) {
				LMSTIC3LogTools.log(Debug.getLogger(), LmsMsg.WarningFehlerEreignisDaten, rdse.getMessage());
			}
		}
	}

	/**
	 * Bestimmt die DAV-Ereigniskodierung.
	 * 
	 * @param event
	 *            {@link TMCEvent}
	 * @param item
	 *            {@link de.bsvrz.dav.daf.main.Data}
	 * @return {@link TMCZustand}
	 */
	private TMCZustand getDavEreignisKodierung(final TMCEvent event, final Data item) {
		TMCZustand ev;
		if (event.getTyp() == TMCEventType.Ereigniscode) {
			// Ein TMC-Ereigniscode von TIC3 kann ein Ereignis- oder
			// Vorhersagecode sein
			// 1. Versuche Ereigniscode (att.rdsEreignisCode)
			ev = new TMCZustand("EreignisCode", event.getCode());
			try {
				ev.setDavDaten(item);
			} catch (final TMCException e1) {
				// kein Ereigniscode (att.rdsEreignisCode)
				// 2. Versuche VorhersageCode (att.rdsVorhersageCode)
				ev = new TMCZustand("VorhersageCode", event.getCode());
				try {
					ev.setDavDaten(item);
				} catch (final TMCException e2) {
					LMSTIC3LogTools.log(Debug.getLogger(), LmsMsg.WarningFehlerEreignisDaten,
							"Der Code '" + event.getCode()
									+ "' kann weder auf einen Ereignis- noch Vorhersagecode abgebildet werden!");
					return null;
				}
			}
		} else if (event.getTyp() == TMCEventType.Empfehlungscode) {
			ev = new TMCZustand("EmpfehlungsCode", event.getCode());
			try {
				ev.setDavDaten(item);
			} catch (final TMCException e) {
				LMSTIC3LogTools.log(Debug.getLogger(), LmsMsg.WarningFehlerEreignisDaten, e.getMessage());
				return null;
			}

		} else {
			LMSTIC3LogTools.log(Debug.getLogger(), LmsMsg.WarningFehlerEreignisDaten, "TMCEventType unbekannt");
			return null;
		}

		return ev;
	}

	private void convertVerortungToDav(final Data item, final ExterneMeldungsDaten tic3daten) {
		final Data ei = item.getItem("VerortungsInformationen");

		final List<TMCLocation> lokationen = tic3daten.getTMCLokationen();

		if (lokationen != null && lokationen.size() > 0) {
			final Array array = ei.getArray("LokationsDaten");
			array.setLength(lokationen.size());

			for (int i = 0; i < lokationen.size(); i++) {
				setLokation(lokationen.get(i), array.getItem(i));
			}
		}

		final Data koordinateItem = ei.getItem("Koordinate");
		final List<TMCKoordinate> koordinaten = tic3daten.getKoordinaten();
		if (koordinaten != null && !koordinaten.isEmpty()) {
			// Die Koordinaten sind in der Meldung in Fahrtrichtung geordnet:
			// Wir nehmen die erste.
			koordinateItem.getItem("TMCLänge").asScaledValue().set(koordinaten.get(0).getLaenge());
			koordinateItem.getItem("TMCBreite").asScaledValue().set(koordinaten.get(0).getBreite());
		} else {
			koordinateItem.getItem("TMCLänge").asUnscaledValue().setText("undefiniert");
			koordinateItem.getItem("TMCBreite").asUnscaledValue().setText("undefiniert");
		}
	}

	private void setLokation(final TMCLocation tmcLocation, final Data item) {
		final Integer primaerLokation = tmcLocation.getPrimaerLokation();
		if (primaerLokation == null) {
			item.getItem("PrimaerLokation").asScaledValue().set(0);
		} else {
			item.getItem("PrimaerLokation").asScaledValue().set(primaerLokation);
		}

		final Integer sekundaerLokation = tmcLocation.getSekundaerLokation();
		if (sekundaerLokation == null) {
			item.getItem("SekundaerLokation").asScaledValue().set(0);
		} else {
			item.getItem("SekundaerLokation").asScaledValue().set(sekundaerLokation);
		}

		item.getItem("TMCRichtung").asUnscaledValue().set(tmcLocation.getTmcRichtung().getId());

		Double entfernung = tmcLocation.getPrimaerEntfernung();
		if (entfernung == null) {
			item.getItem("StartOffset").asUnscaledValue().set(0);
		} else {
			item.getItem("StartOffset").asScaledValue().set(entfernung);
		}

		entfernung = tmcLocation.getSekundaerEntfernung();
		if (entfernung == null) {
			item.getItem("EndOffset").asUnscaledValue().set(0);
		} else {
			item.getItem("EndOffset").asScaledValue().set(entfernung);
		}
	}

	@Override
	public boolean hasTmcDaten() {
		final List<TMCLocation> lokationen = daten.getTMCLokationen();
		final List<TMCKoordinate> koordinaten = daten.getKoordinaten();

		return (lokationen != null && lokationen.size() > 0) || (koordinaten != null && koordinaten.size() > 0);
	}
}
